import re
from io import BytesIO

import numpy as np
import pandas as pd

from wbportfolio.models import Product


def file_name_parse(file_name):
    isin = re.findall("([A-Z]{2}[A-Z0-9]{9}[0-9]{1})", file_name)

    if len(isin) != 1:
        raise ValueError("Not exactly 1 isin found in the filename")

    return {"isin": isin[0]}


# TODO: finish it
_fields_map = {
    "underlying_instrument": "underlying_instrument",
    "transaction_subtype": "transaction_subtype",
    "DATE": "transaction_date",
    "INSTRUMENT CCY": "currency__key",
    "TOTAL QUANTITY": "shares",
    "GROSS PRICE": "price_gross",
    "NET PRICE": "price",
    "currency_fx_rate": "currency_fx_rate",
    "NET AMOUNT": "total_value",
    "N°": "external_id",
}


def parse(import_source):
    data = list()
    parts = file_name_parse(import_source.file.name)
    product = Product.objects.get(isin=parts["isin"])

    df = pd.read_excel(BytesIO(import_source.file.read()), engine="openpyxl", sheet_name="Transactions")

    def _get_exchange(identifier):
        ticker, exchange, _ = identifier.split(" ")
        if exchange:
            return {"bbg_exchange_codes": exchange}

    def _get_underlying_instrument(row):
        if row["EVENT"] == "AMC CASH":
            return product.id
        else:
            ticker, exchange, _ = row["IDENTIFIER"].split(" ")
            return {
                "ticker": ticker,
                "exchange": {"bbg_exchange_codes": exchange},
                "name": row["NAME"],
                "currency__key": row["currency__key"],
                "instrument_type": "equity",
            }

    xx, yy = np.where(df == "N°")
    df = df.iloc[xx[0] :, yy[0] :]
    df = df.rename(columns=df.iloc[0]).drop(df.index[0]).dropna(how="all")
    df["transaction_subtype"] = None
    df = df.rename(columns={"INSTRUMENT CCY  ": "INSTRUMENT CCY"})
    df = df.rename(columns=_fields_map)

    df["transaction_date"] = pd.to_datetime(df["transaction_date"], dayfirst=True).dt.strftime("%Y-%m-%d")
    df["shares"] = -1 * df["shares"]

    df.loc[(df["EVENT"] == "BUY") & (df["ASSET CLASS"] == "SHARE"), "transaction_subtype"] = "BUY"
    df.loc[(df["EVENT"] == "SELL") & (df["ASSET CLASS"] == "SHARE"), "transaction_subtype"] = "SELL"
    df.loc[(df["EVENT"] == "AMC CASH") & (df["shares"] < 0), "transaction_subtype"] = "REDEMPTION"
    df.loc[(df["EVENT"] == "AMC CASH") & (df["shares"] > 0), "transaction_subtype"] = "SUBSCRIPTION"

    df = df[df["transaction_subtype"].notnull()]
    df["underlying_instrument"] = df.apply(lambda x: _get_underlying_instrument(x), axis=1)
    df["exchange"] = df["IDENTIFIER"].apply(lambda x: _get_exchange(x))
    df["currency_fx_rate"] = df["BOOKING FX"] * df["BASE CCY FX"]

    df = df.drop(df.columns.difference(_fields_map.values()), axis=1)
    df["portfolio"] = product.primary_portfolio.id

    customer_trade_index_mask = (df["transaction_subtype"] == "REDEMPTION") | (
        df["transaction_subtype"] == "SUBSCRIPTION"
    )
    df.loc[customer_trade_index_mask, "shares"] = df.loc[customer_trade_index_mask, "shares"] / product.share_price
    for field in ["price", "price_gross"]:
        df.loc[customer_trade_index_mask, field] = df.loc[customer_trade_index_mask, field] * product.share_price
    df["bank"] = "Leonteq Cash Transfer"
    data = df.to_dict("records")

    return {"data": data}



from django.contrib import admin
from wbfdm.admin import InstrumentModelAdmin

from wbportfolio.models import Index


@admin.register(Index)
class IndexAdmin(InstrumentModelAdmin):
    fieldsets = (
        ("Instrument Information", InstrumentModelAdmin.fieldsets[0][1]),
        (
            "Index Information",
            {"fields": (("net_asset_value_computation_method_path", "order_routing_custodian_adapter"),)},
        ),
    )



import datetime
import logging
import re

import numpy as np
import pandas as pd

from wbportfolio.import_export.utils import convert_string_to_number

logger = logging.getLogger("importers.parsers.jpmorgan.index")


def file_name_parse(file_name):
    dates = re.findall("([0-9]{8})", file_name)

    if len(dates) != 1:
        raise ValueError("Not exactly 1 date found in the filename")
    return {"valuation_date": datetime.datetime.strptime(dates[0], "%Y%m%d").date()}


def parse(import_source):
    # Load files into a CSV DictReader
    df = pd.read_csv(import_source.file, encoding="utf-16", delimiter=",")
    df = df.replace([np.inf, -np.inf, np.nan], None)

    # Iterate through the CSV File and parse the data into a list
    data = list()
    for price_data in df.to_dict("records"):
        date = datetime.datetime.strptime(price_data["Date"], "%d-%b-%y")

        data.append(
            {
                "instrument": {"instrument_type": "product", "isin": price_data["ISIN"]},
                "date": date.strftime("%Y-%m-%d"),
                "net_value": round(convert_string_to_number(price_data["Indicative_MID"]), 4),
                "calculated": False,
            }
        )

    return {"data": data}



from unittest.mock import patch

import pandas as pd
import pytest
from faker import Faker
from psycopg.types.range import NumericRange
from wbcompliance.factories.risk_management import RuleThresholdFactory
from wbfdm.analysis.esg.esg_analysis import DataLoader

from wbportfolio.risk_management.backends.esg_aggregation_portfolio import (
    RuleBackend as ESGAggregationPortfolioBackend,
)
from wbportfolio.tests.models.utils import PortfolioTestMixin

fake = Faker()


@pytest.mark.django_db
class TestEsgAggregationPortfolioRuleModel(PortfolioTestMixin):
    @patch.object(DataLoader, "compute")
    def test_eval(
        self,
        mock_fct,
        weekday,
        asset_position_factory,
        portfolio,
    ):
        parameters = {"esg_aggregation": "GHG_EMISSIONS_SCOPE_1"}
        backend = ESGAggregationPortfolioBackend(
            weekday,
            portfolio,
            parameters,
            [RuleThresholdFactory.create(range=NumericRange(lower=0.02, upper=0.03))],  # type: ignore
        )

        a1 = asset_position_factory.create(
            date=weekday,
            portfolio=portfolio,
        )  # Breached position

        a2 = asset_position_factory.create(
            date=weekday,
            portfolio=portfolio,
        )
        mock_fct.return_value = pd.Series(index=[a1.underlying_quote.id, a2.underlying_quote.id], data=[0.01, 0.025])
        incidents = list(backend.check_rule())
        assert len(incidents) == 1
        incident = incidents[0]
        assert incident.breached_object == a2.underlying_quote



from .adjustments import AdjustmentEndpointConfig, AdjustmentEquityEndpointConfig
from .assets import (
    AssetPositionEndpointConfig,
    AssetPositionEquityEndpointConfig,
    AssetPositionIndexEndpointConfig,
    AssetPositionInstrumentEndpointConfig,
    AssetPositionPortfolioEndpointConfig,
    AssetPositionProductGroupEndpointConfig,
    AssetPositionUnderlyingInstrumentChartEndpointConfig,
    CashPositionPortfolioEndpointConfig,
    CompositionModelPortfolioPandasEndpointConfig,
    ContributorPortfolioChartEndpointConfig,
    DistributionChartEndpointConfig,
    DistributionTableEndpointConfig,
)
from .claim import (
    ConsolidatedTradeSummaryEndpointConfig,
    ConsolidatedTradeSummaryDistributionChartEndpointConfig,
    CumulativeNNMChartEndpointConfig,
    ClaimAccountEndpointConfig,
    ClaimEndpointConfig,
    ClaimEntryEndpointConfig,
    ClaimProductEndpointConfig,
    ClaimTradeEndpointConfig,
    NegativeTermimalAccountPerProductEndpointConfig,
    ProfitAndLossPandasEndpointConfig,
)

from .custodians import CustodianEndpointConfig
from .fees import (
    FeesAggregatedProductPandasEndpointConfig,
    FeesProductEndpointConfig,
)
from .fees import FeesAggregatedProductPandasEndpointConfig, FeesProductEndpointConfig, FeeEndpointConfig

from .portfolios import (
    PortfolioEndpointConfig,
    PortfolioPortfolioThroughModelEndpointConfig,
    PortfolioTreeGraphChartEndpointConfig,
    TopDownPortfolioCompositionPandasEndpointConfig
)
from .positions import (
    AggregatedAssetPositionLiquidityEndpointConfig,
    AssetPositionPandasEndpointConfig,
)
from .product_groups import ProductGroupEndpointConfig
from .product_performance import (
    PerformanceComparisonEndpointConfig,
    PerformancePandasEndpointConfig,
    ProductPerformanceNetNewMoneyEndpointConfig,
)
from .products import (
    ProductCustomerEndpointConfig,
    ProductPerformanceFeesEndpointConfig,
    InstrumentPriceAUMDataEndpointConfig,
    NominalProductEndpointConfig,
    AUMProductEndpointConfig,
)
from .roles import PortfolioRoleInstrumentEndpointConfig
from .trades import (
    CustodianDistributionInstrumentEndpointConfig,
    TradeEndpointConfig,
    TradeInstrumentEndpointConfig,
    TradePortfolioEndpointConfig,
    SubscriptionRedemptionEndpointConfig,
)
from .portfolio_relationship import (
    PortfolioInstrumentPreferredClassificationThroughEndpointConfig,
    InstrumentPortfolioThroughPortfolioModelEndpointConfig,
)
from .esg import ESGMetricAggregationPortfolioPandasEndpointConfig
from .reconciliations import AccountReconciliationLineEndpointViewConfig, AccountReconciliationEndpointViewConfig
from .rebalancing import RebalancerEndpointConfig



from decimal import Decimal

from django.core.exceptions import ObjectDoesNotExist
from wbfdm.models import InstrumentPrice

from wbportfolio.pms.typing import Portfolio, Position
from wbportfolio.rebalancing.base import AbstractRebalancingModel
from wbportfolio.rebalancing.decorators import register


@register("Composite Rebalancing")
class CompositeRebalancing(AbstractRebalancingModel):
    @property
    def base_assets(self) -> dict[int, Decimal]:
        """
        Return a dictionary representation (instrument_id: target weight) of this order proposal
        Returns:
            A dictionary representation

        """
        try:
            latest_order_proposal = self.portfolio.order_proposals.filter(
                status="CONFIRMED", trade_date__lt=self.trade_date
            ).latest("trade_date")
            return {
                v["underlying_instrument"]: v["target_weight"]
                for v in latest_order_proposal.get_orders().values("underlying_instrument", "target_weight")
            }
        except ObjectDoesNotExist:
            return dict()

    def is_valid(self) -> bool:
        return (
            len(self.base_assets.keys()) > 0
            and InstrumentPrice.objects.filter(date=self.trade_date, instrument__in=self.base_assets.keys()).exists()
        )

    def get_target_portfolio(self) -> Portfolio:
        positions = []
        for underlying_instrument, weighting in self.base_assets.items():
            positions.append(
                Position(underlying_instrument=underlying_instrument, weighting=weighting, date=self.trade_date)
            )
        return Portfolio(positions=tuple(positions))



import pandas as pd
from pandas.tseries.offsets import BDay

from .utils import file_name_parse_isin

FIELD_MAP = {
    "Date": "fee_date",
    "Manag. Fees Natixis": "ISSUER",
    "Manag. Fees Client": "MANAGEMENT",
    "Perf fees amount": "PERFORMANCE",
    # "Currency": "currency"
}


def parse(import_source):
    # Parse the Parts of the filename into the different parts
    parts = file_name_parse_isin(import_source.file.name)
    # Get the valuation date and investment from the parts list
    product = parts["product"]

    df = pd.read_csv(import_source.file, encoding="utf-8", delimiter=";")
    df = df.rename(columns=FIELD_MAP)
    df = df.drop(columns=df.columns.difference(FIELD_MAP.values()))
    df["fee_date"] = pd.to_datetime(df["fee_date"], dayfirst=True)

    # Switch the weeekend day fees to the next monday
    df["fee_date"] = df["fee_date"].apply(lambda x: x + BDay(1) if x.weekday() in [5, 6] else x)

    # Ensure float columns are number
    for col in ["MANAGEMENT", "ISSUER", "PERFORMANCE"]:
        df[col] = df[col].astype("str").str.replace(" ", "").astype("float")

    # Groupby and sum similar fees (e.g. Monday)
    df = df.groupby("fee_date").sum().reset_index()
    df = pd.melt(
        df,
        id_vars=["fee_date"],
        value_vars=["MANAGEMENT", "ISSUER", "PERFORMANCE"],
        var_name="transaction_subtype",
        value_name="total_value",
    )
    df = df[df["total_value"] != 0]

    df["product"] = [product] * df.shape[0]
    df["fee_date"] = df["fee_date"].dt.strftime("%Y-%m-%d")
    df["calculated"] = False
    df["total_value_gross"] = df["total_value"]
    return {"data": df.to_dict("records")}



from datetime import date, timedelta

import pytest
from pandas._libs.tslibs.offsets import BDay

from wbportfolio.analysis.claims import ConsolidatedTradeSummary
from wbportfolio.models import Claim


@pytest.mark.django_db
class TestConsolidatedTradeSummary:
    def test_base(self, claim_factory, customer_trade_factory, product, instrument_price_factory):
        d1 = date(2024, 1, 1)
        d2 = date(2025, 1, 1)
        p1 = instrument_price_factory.create(date=d1, instrument=product)  # noqa
        p2 = instrument_price_factory.create(date=d2, instrument=product)
        t1 = customer_trade_factory.create(
            shares=1000, transaction_date=d1, underlying_instrument=product, portfolio=product.primary_portfolio
        )
        t2 = customer_trade_factory.create(
            shares=500, transaction_date=d2, underlying_instrument=product, portfolio=product.primary_portfolio
        )

        c0 = claim_factory.create()  # noqa
        c1 = claim_factory.create(shares=1000, status="APPROVED", trade=t1, date=d2, product=product)
        c2 = claim_factory.create(shares=500, status="APPROVED", trade=t2, date=d1, product=product)
        cts = ConsolidatedTradeSummary(Claim.objects.all(), d1, d2, "account", "account__title")

        valid_date = dict(cts.queryset.values_list("id", "valid_date"))
        assert valid_date == {c1.id: d2, c2.id: d2}

        date_considered = dict(cts.queryset.values_list("id", "date_considered"))
        assert date_considered == {c1.id: d2 + timedelta(days=1), c2.id: d2 + timedelta(days=1)}

        aum = dict(cts.queryset.values_list("id", "aum"))
        assert aum == {c1.id: p2.net_value * c1.shares, c2.id: p2.net_value * c2.shares}

    def test_get_aum_df(self, claim_factory, product, instrument_price_factory, customer_trade_factory):
        t1 = customer_trade_factory.create(
            shares=10,
            transaction_date=date(2024, 12, 31),
            underlying_instrument=product,
            portfolio=product.primary_portfolio,
        )
        t2 = customer_trade_factory.create(
            shares=100,
            transaction_date=date(2025, 2, 3),
            underlying_instrument=product,
            portfolio=product.primary_portfolio,
        )
        c1 = claim_factory.create(trade=t1, status="APPROVED")
        c2 = claim_factory.create(trade=t2, status="APPROVED", account=c1.account)
        d1 = date(2025, 1, 1)
        d2 = date(2025, 3, 3)
        product.prices.all().delete()
        p1 = instrument_price_factory.create(
            date=(d1 - BDay(7)).date(), instrument=product
        )  # we test that the CTS uses the earliest date
        p2 = instrument_price_factory.create(date=c2.date, instrument=product)  # noqa
        p3 = instrument_price_factory.create(date=d2, instrument=product)

        cts = ConsolidatedTradeSummary(Claim.objects.all(), d1, d2, "account", "account__title")
        aum_start = float(round(p1.net_value * c1.shares))
        aum_end = float(round(p3.net_value * (c1.shares + c2.shares)))
        assert cts.get_aum_df().to_dict(orient="index") == {
            c1.account.id: {
                "sum_shares_start": c1.shares,
                "sum_shares_end": c1.shares + c2.shares,
                "sum_aum_start": aum_start,
                "sum_aum_end": aum_end,
                "sum_shares_diff": c2.shares,
                "sum_shares_perf": (c1.shares + c2.shares) / c1.shares - 1,
                "sum_aum_diff": aum_end - aum_start,
                "sum_aum_perf": aum_end / aum_start - 1,
            }
        }

        p0 = instrument_price_factory.create(date=c1.date, instrument=product)  # noqa

        assert cts.get_nnm_df().to_dict(orient="index") == {
            c1.account.id: {
                "sum_nnm_2025-02": float(round(c2.shares * p2.net_value)),
                "sum_nnm_total": float(round(c2.shares * p2.net_value)),
            }
        }, "Ensure the first claim that is considered in t+1 (on the CTS lower bound range) is excluded from the NNM"



from abc import ABC, abstractmethod

from wbportfolio.order_routing import ExecutionStatus
from wbportfolio.pms.typing import Order

class BaseCustodianAdapter(ABC):

    def __init__(self, isin: str, **identifiers):
        self.isin = isin

    @property
    def errors(self):
        if not hasattr(self, '_errors'):
            raise ValueError("is_valid needs to call before accessing errors")
        return
    @abstractmethod
    def authenticate(self) -> bool:
        """
        Authenticate or renew tokens with the custodian API.
        Raises an exception if authentication fails.
        """
        pass

    @abstractmethod
    def is_valid(self) -> bool:
        """
        Check whether the given isin is valid and can be rebalanced
        """
        pass

    @abstractmethod
    def serialize_orders(self, orders: list[Order]) -> list[dict[str, str]]:
        pass

    @abstractmethod
    def deserialize_items(self, items: list[dict[str, str]]) -> list[Order]:
        pass

    @abstractmethod
    def get_rebalance_status(self) -> tuple[ExecutionStatus, str]:
        """
        Return the rebalance status as a string (in the custodian format)
        """
        pass

    @abstractmethod
    def submit_rebalancing(self, items: list[dict[str, str]], as_draft: bool = True) -> tuple[list[dict[str, str]], str]:
        """
        Submit a rebalance order for the certificate.
        """
        pass

    @abstractmethod
    def cancel_current_rebalancing(self) -> bool:
        """
        Cancel an existing rebalance order identified by ISIN.
        """
        pass

    @abstractmethod
    def get_current_rebalancing(self) -> list[dict[str, str]]:
        """
        Fetch the current rebalance request details for a certificate.
        """
        pass



from django.contrib import admin

from wbportfolio.models import Register


@admin.register(Register)
class RegisterModelAdmin(admin.ModelAdmin):
    search_fields = [
        "register_reference",
        "register_name_1",
        "register_name_2",
        "custodian_reference",
        "custodian_name_1",
        "custodian_name_2",
        "outlet_reference",
        "outlet_name",
    ]
    list_display = ("register_reference", "global_register_reference", "register_name_1", "register_name_2")



from datetime import date

from django.dispatch import receiver
from rest_framework.reverse import reverse
from wbcore.contrib.directory.serializers import (
    CompanyModelSerializer,
    EntryModelSerializer,
    PersonModelSerializer,
)
from wbcore.filters.defaults import current_quarter_date_start
from wbcore.signals import add_instance_additional_resource
from wbcrm.serializers.accounts import AccountModelSerializer
from wbfdm.serializers import InstrumentModelSerializer

from wbportfolio.models import PortfolioRole, Trade
from wbportfolio.serializers.products import ProductModelSerializer


@receiver(add_instance_additional_resource, sender=InstrumentModelSerializer)
@receiver(add_instance_additional_resource, sender=ProductModelSerializer)
def add_instrument_serializer_resources(sender, serializer, instance, request, user, **kwargs):
    additional_resources = dict()

    has_assets = any([i.assets.exists() for i in instance.get_descendants(include_self=True)])
    if PortfolioRole.is_portfolio_manager(user.profile) and has_assets:
        additional_resources["assets"] = reverse(
            "wbportfolio:instrument-asset-list", args=[instance.id], request=request
        )
        additional_resources["assetschart"] = reverse(
            "wbportfolio:instrument-assetpositionchart-list", args=[instance.id], request=request
        )

    if instance.portfolios.exists():
        additional_resources["distribution_chart"] = reverse(
            "wbportfolio:portfolio-distributionchart-list", args=[instance.portfolio.id], request=request
        )
        additional_resources["distribution_table"] = reverse(
            "wbportfolio:portfolio-distributiontable-list", args=[instance.portfolio.id], request=request
        )

    if user.is_internal or user.is_superuser:
        if Trade.valid_customer_trade_objects.filter(underlying_instrument=instance).exists():
            additional_resources["instrument_subscriptionsredemptions"] = reverse(
                "wbportfolio:instrument-subscriptionredemption-list", args=[instance.id], request=request
            )
        if (
            Trade.objects.filter(underlying_instrument__in=instance.get_descendants(include_self=True))
            .exclude(transaction_subtype__in=[Trade.Type.REDEMPTION, Trade.Type.SUBSCRIPTION])
            .exists()
        ):
            additional_resources["instrument_trades"] = (
                f"{reverse('wbportfolio:instrument-trade-list', args=[instance.id], request=request)}?is_customer_trade=False"
            )

    if PortfolioRole.is_manager(user.profile):
        additional_resources["portfoliorole"] = reverse(
            "wbportfolio:instrument-portfoliorole-list", args=[instance.id], request=request
        )
    return additional_resources


@receiver(add_instance_additional_resource, sender=InstrumentModelSerializer)
@receiver(add_instance_additional_resource, sender=ProductModelSerializer)
def asset_portfolio_resources(sender, serializer, instance, request, user, **kwargs):
    additional_resources = dict()
    portfolio = instance.portfolio

    if portfolio and portfolio.assets.exists():
        if PortfolioRole.is_analyst(user.profile, portfolio=portfolio):
            additional_resources["portfolio_positions"] = reverse(
                "wbportfolio:portfolio-asset-list", args=[portfolio.id], request=request
            )
            additional_resources["portfolio_positions_contributors"] = reverse(
                "wbportfolio:portfolio-contributor-list",
                args=[portfolio.id],
                request=request,
            )
    return additional_resources


@receiver(add_instance_additional_resource, sender=InstrumentModelSerializer)
@receiver(add_instance_additional_resource, sender=ProductModelSerializer)
def transactions_resources(sender, serializer, instance, request, user, **kwargs):
    additional_resources = dict()
    portfolio = instance.primary_portfolio
    if portfolio:
        if portfolio.trades.exists() and user.is_internal:
            additional_resources["portfolio_subscriptionsredemptions"] = (
                f"{reverse('wbportfolio:portfolio-trade-list', args=[portfolio.id], request=request)}?is_customer_trade=True"
            )
            additional_resources["portfolio_trades"] = (
                f"{reverse('wbportfolio:portfolio-trade-list', args=[portfolio.id], request=request)}?is_customer_trade=False"
            )

    return additional_resources


@receiver(add_instance_additional_resource, sender=ProductModelSerializer)
def product_resources(sender, serializer, instance, request, user, **kwargs):
    additional_resources = dict()
    if user.is_internal or user.is_superuser:
        additional_resources["custodiandistribution"] = reverse(
            "wbportfolio:instrument-custodiandistribution-list",
            args=[instance.id],
            request=request,
        )
        additional_resources["customerdistribution"] = reverse(
            "wbportfolio:instrument-customerdistribution-list",
            args=[instance.id],
            request=request,
        )
        additional_resources["tradechart"] = reverse(
            "wbportfolio:product-nominalchart-list",
            args=[instance.id],
            request=request,
        )
        additional_resources["aumchart"] = reverse(
            "wbportfolio:product-aumchart-list",
            args=[instance.id],
            request=request,
        )
    if instance.fees.exists() and PortfolioRole.is_portfolio_manager(user.profile, instrument=instance):
        additional_resources["product_fees"] = reverse(
            "wbportfolio:product-fees-list", args=[instance.id], request=request
        )
        additional_resources["product_aggregatedfees"] = reverse(
            "wbportfolio:product-feesaggregated-list", args=[instance.id], request=request
        )

    return additional_resources


@receiver(add_instance_additional_resource, sender=CompanyModelSerializer)
@receiver(add_instance_additional_resource, sender=PersonModelSerializer)
@receiver(add_instance_additional_resource, sender=EntryModelSerializer)
def claim_adding_additional_resource(sender, serializer, instance, request, user, **kwargs):
    start = current_quarter_date_start()
    return {
        "claims": reverse("wbportfolio:entry-claim-list", args=[instance.id], request=request),
        "aum": f"{reverse('wbportfolio:aumtable-list', args=[], request=request)}?group_by=PRODUCT&account_owner={instance.id}&date={start:%Y-%m-%d},{date.today():%Y-%m-%d}",
    }


@receiver(add_instance_additional_resource, sender=AccountModelSerializer)
def account_reconciliation_resource(sender, serializer, instance, request, user, **kwargs):
    return {
        "wbportfolio-accountreconciliation": f"{reverse('wbportfolio:accountreconciliation-list', request=request)}?account={instance.id}",
        "wbportfolio-reconcile-account": f"{reverse('wbportfolio:accountreconciliation-list', request=request)}?account={instance.id}",
    }



from typing import Optional

from wbcore.contrib.authentication.models import User
from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.directory.models import Person
from wbcore.enums import Unit
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import Display
from wbcore.metadata.configs.display.instance_display.styles import Style
from wbcore.metadata.configs.display.instance_display.utils import repeat
from wbcore.metadata.configs.display.view_config import DisplayViewConfig
from wbcore.metadata.configs.display.windows import Window

from wbportfolio.models.transactions.claim import Claim


class ClaimDisplayConfig(DisplayViewConfig):
    LEGENDS = [
        dp.Legend(
            key="status",
            items=[
                dp.LegendItem(
                    icon=WBColor.YELLOW_LIGHT.value,
                    label=Claim.Status.DRAFT.label,
                    value=Claim.Status.DRAFT.name,
                ),
                dp.LegendItem(
                    icon=WBColor.YELLOW.value,
                    label=Claim.Status.AUTO_MATCHED.label,
                    value=Claim.Status.AUTO_MATCHED.name,
                ),
                dp.LegendItem(
                    icon=WBColor.BLUE_LIGHT.value,
                    label=Claim.Status.PENDING.label,
                    value=Claim.Status.PENDING.name,
                ),
                dp.LegendItem(
                    icon=WBColor.GREEN_LIGHT.value,
                    label=Claim.Status.APPROVED.label,
                    value=Claim.Status.APPROVED.name,
                ),
                dp.LegendItem(
                    icon=WBColor.GREY.value,
                    label=Claim.Status.WITHDRAWN.label,
                    value=Claim.Status.WITHDRAWN.name,
                ),
            ],
        )
    ]
    LIST_FORMATTING = [
        dp.Formatting(
            column="status",
            formatting_rules=[
                dp.FormattingRule(
                    style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                    condition=("==", Claim.Status.APPROVED.name),
                ),
                dp.FormattingRule(
                    style={"backgroundColor": WBColor.YELLOW.value},
                    condition=("==", Claim.Status.AUTO_MATCHED.name),
                ),
                dp.FormattingRule(
                    style={"backgroundColor": WBColor.BLUE_LIGHT.value},
                    condition=("==", Claim.Status.PENDING.name),
                ),
                dp.FormattingRule(
                    style={"backgroundColor": WBColor.GREY.value},
                    condition=("==", Claim.Status.WITHDRAWN.name),
                ),
                dp.FormattingRule(
                    style={"backgroundColor": WBColor.YELLOW_LIGHT.value},
                    condition=("==", Claim.Status.DRAFT.name),
                ),
            ],
        )
    ]

    def get_list_display(self):
        user: User = self.request.user
        profile: Person = user.profile
        fields = [
            dp.Field(key="date", label="Date", width=Unit.PIXEL(100)),
            (
                dp.Field(key="product", label="Product", width=Unit.PIXEL(300))
                if "product_id" not in self.view.kwargs
                else ()
            ),
            dp.Field(key="claimant", label="Claimant", width=Unit.PIXEL(250)),
            (
                dp.Field(key="account", label="Account", width=Unit.PIXEL(150))
                if "account_id" not in self.view.kwargs
                else ()
            ),
            dp.Field(key="trade_comment", label="Comment", width=Unit.PIXEL(250)),
            dp.Field(
                key="shares",
                label="Shares",
                formatting_rules=[
                    dp.FormattingRule(
                        style={"color": WBColor.RED_DARK.value, "fontWeight": "bold"},
                        condition=("<", 0),
                    )
                ],
                width=Unit.PIXEL(100),
            ),
            dp.Field(key="last_nav", label="NAV", width=Unit.PIXEL(100)),
            dp.Field(key="total_value", label="Value", width=Unit.PIXEL(100)),
            dp.Field(key="total_value_usd", label="Value (USD)", width=Unit.PIXEL(100)),
            dp.Field(key="bank", label="Bank", width=Unit.PIXEL(100)),
        ]
        if profile.is_internal or user.is_superuser:
            fields.extend(
                [
                    dp.Field(key="creator", label="Creator", width=Unit.PIXEL(150)),
                    (
                        dp.Field(key="trade", label="Trade", width=Unit.PIXEL(250))
                        if "trade_id" not in self.view.kwargs
                        else ()
                    ),
                ]
            )
        fields.append(dp.Field(key="reference", label="Reference", width=Unit.PIXEL(150)))
        return dp.ListDisplay(
            fields=fields,
            formatting=ClaimDisplayConfig.LIST_FORMATTING,
            legends=ClaimDisplayConfig.LEGENDS,
        )

    def get_window(self) -> Window | None:
        if self.instance or self.new_mode:
            if self.profile.is_internal or self.user.is_superuser:
                return Window(max_width=500, min_width=500, max_height=900, min_height=900)
            return Window(max_width=500, min_width=500, max_height=750, min_height=750)
        return None

    def get_instance_display(self) -> Display:
        fields = [
            ["trade_type", "trade_type"],
            ["product", "product"],
            ["as_shares", "quantity"],
            ["bank", "bank"],
            ["date", "date"],
            ["account", "account"],
            ["reference", "reference"],
        ]
        if "entry_id" not in self.view.kwargs:
            fields.append(["claimant", "claimant"])
        # For client we hide the status and trade as it does not concern them
        if self.profile.is_internal or self.user.is_superuser:
            fields.insert(0, ["status", "status"])
            if "trade_id" not in self.view.kwargs:
                fields.insert(6, ["trade", "trade"])

        return dp.Display(
            pages=[
                dp.Page(
                    layouts={
                        "default": dp.Layout(
                            grid_template_areas=fields,
                            grid_template_columns=[repeat(2, Style.minmax(Style.px(200), Style.px(250)))],
                        )
                    }
                )
            ]
        )


class ConsolidatedTradeSummaryDisplayConfig(DisplayViewConfig):
    DEFAULT_COL_WIDTH = 100

    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = [
            dp.Field(key="title", label="Title", pinned="left", width=250),
            dp.Field(
                key="initial_investment_date", label="Initial Investment", pinned="left", width=self.DEFAULT_COL_WIDTH
            ),
            dp.Field(key="aum_sparkline", label="NNM Chart", pinned="left"),
            dp.Field(
                key="shares",
                label="Outstanding Shares",
                open_by_default=False,
                children=[
                    dp.Field(key="sum_shares_start", label="Start", width=self.DEFAULT_COL_WIDTH, show="open"),
                    dp.Field(key="sum_shares_end", label="End", width=self.DEFAULT_COL_WIDTH),
                    dp.Field(key="sum_shares_diff", label="Δ", width=self.DEFAULT_COL_WIDTH, show="open"),
                    dp.Field(key="sum_shares_perf", label="%", width=self.DEFAULT_COL_WIDTH),
                ],
                marry_children=True,
            ),
            dp.Field(
                key="aum_stat",
                label="AUM",
                open_by_default=False,
                children=[
                    dp.Field(key="sum_aum_start", label="Start", width=self.DEFAULT_COL_WIDTH, show="open"),
                    dp.Field(key="sum_aum_end", label="End", width=self.DEFAULT_COL_WIDTH),
                    dp.Field(key="sum_aum_diff", label="Δ", width=self.DEFAULT_COL_WIDTH, show="open"),
                    dp.Field(key="sum_aum_perf", label="%", width=self.DEFAULT_COL_WIDTH),
                ],
                marry_children=True,
            ),
            dp.Field(
                key="nnm_stat",
                label="NNM",
                open_by_default=False,
                children=[
                    *map(
                        lambda monthly_nnm_column: dp.Field(
                            key=monthly_nnm_column[0],
                            label=monthly_nnm_column[1],
                            show="open",
                            width=self.DEFAULT_COL_WIDTH,
                        ),
                        self.view.nnm_monthly_columns,
                    ),
                    dp.Field(
                        key="sum_nnm_total",
                        label="Total",
                        width=self.DEFAULT_COL_WIDTH,
                    ),
                    dp.Field(
                        key="sum_nnm_perf",
                        label="%",
                        width=self.DEFAULT_COL_WIDTH,
                    ),
                ],
                marry_children=True,
            ),
            dp.Field(
                key="performance_stat",
                label="Performance",
                open_by_default=False,
                children=[
                    dp.Field(
                        key="total_performance",
                        label="Total",
                        width=self.DEFAULT_COL_WIDTH,
                    ),
                    dp.Field(
                        key="total_performance_perf",
                        label="%",
                        width=self.DEFAULT_COL_WIDTH,
                    ),
                ],
                marry_children=True,
            ),
        ]

        if self.view.commission_type_columns:
            fields.append(
                dp.Field(
                    key="rebate_stat",
                    label="Rebate",
                    open_by_default=False,
                    children=[
                        *map(
                            lambda commission_type_column: dp.Field(
                                key=commission_type_column[0],
                                label=commission_type_column[1],
                                show="open",
                                width=self.DEFAULT_COL_WIDTH,
                            ),
                            self.view.commission_type_columns,
                        ),
                        dp.Field(
                            key="rebate_total",
                            label="Total",
                            width=self.DEFAULT_COL_WIDTH,
                        ),
                    ],
                    marry_children=True,
                )
            )

        return dp.ListDisplay(fields=fields)


class ProfitAndLossPandasDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="title", label="Title"),
                dp.Field(key="unrealized_pnl", label="Realized P&L"),
                dp.Field(key="realized_pnl", label="Unrealized P&L"),
                dp.Field(key="total_pnl", label="Total P&L"),
                dp.Field(key="total_invested", label="Total Invested"),
                dp.Field(key="performance", label="Performance"),
            ]
        )


class NegativeTermimalAccountPerProductDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="account_repr", label="Account"),
                dp.Field(key="product_repr", label="Product"),
                dp.Field(key="sum_shares", label="Total Shares"),
            ]
        )



from decimal import Decimal

from wbportfolio.models.utils import get_adjusted_shares


def test_get_adjusted_shares():
    assert get_adjusted_shares(Decimal("150"), Decimal("100"), Decimal("200")) == Decimal("75")



from contextlib import suppress
from typing import TYPE_CHECKING

from django.db import models
from django.db.models import Q
from django.dispatch import receiver
from wbcore.signals import pre_merge
from wbfdm.models import Instrument

if TYPE_CHECKING:
    from wbfdm.models import Exchange

    from wbportfolio.models.portfolio import Portfolio


class PortfolioBankAccountThroughModel(models.Model):
    class PortfolioBankAccountType(models.TextChoices):
        CASH = "CASH", "Cash"
        SECURITIES = "SECURITIES", "Securities"

    portfolio = models.ForeignKey(
        to="wbportfolio.Portfolio", on_delete=models.CASCADE, related_name="bank_account_through"
    )
    bank_account = models.ForeignKey(
        to="directory.BankingContact", on_delete=models.CASCADE, related_name="portfolio_through"
    )

    portfolio_bank_account_type = models.CharField(
        max_length=255, choices=PortfolioBankAccountType.choices, default=PortfolioBankAccountType.CASH
    )

    def __str__(self) -> str:
        return f"{self.bank_account}: {self.portfolio} ({self.PortfolioBankAccountType[self.portfolio_bank_account_type].label})"

    class Meta:
        verbose_name = "Portfolio Bank Account"
        verbose_name_plural = "Portfolio Bank Accounts"

        constraints = [
            # Each portfolio and bank account can only be connected once
            models.UniqueConstraint(
                fields=["portfolio", "bank_account"],
                name="unique_portfolio_bank_account",
            ),
            # Each portfolio can only have one bank account to hold the securities
            models.UniqueConstraint(
                fields=["portfolio"],
                name="unique_portfolio_bank_account_type",
                condition=Q(portfolio_bank_account_type="SECURITIES"),
            ),
        ]


class InstrumentPortfolioThroughModel(models.Model):
    instrument = models.ForeignKey(
        "wbfdm.Instrument",
        on_delete=models.CASCADE,
        related_name="through_portfolios",
    )
    portfolio = models.ForeignKey(
        "wbportfolio.Portfolio", on_delete=models.CASCADE, related_name="through_instruments"
    )

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=["instrument"], name="unique_instrument"),
            models.UniqueConstraint(fields=["instrument", "portfolio"], name="unique_portfolio_relationship"),
        ]

    def __str__(self) -> str:
        return f"{self.instrument} - {self.portfolio}"

    @classmethod
    def get_portfolio(cls, instrument):
        with suppress(InstrumentPortfolioThroughModel.DoesNotExist):
            return InstrumentPortfolioThroughModel.objects.get(instrument=instrument).portfolio

    @classmethod
    def get_primary_portfolio(cls, instrument):
        if portfolio := cls.get_portfolio(instrument):
            if primary_portfolio := portfolio.primary_portfolio:
                return primary_portfolio
            return portfolio


class PortfolioInstrumentPreferredClassificationThroughModel(models.Model):
    portfolio = models.ForeignKey(
        "wbportfolio.Portfolio", on_delete=models.CASCADE, related_name="preferred_classification_instruments"
    )
    instrument = models.ForeignKey(
        "wbfdm.Instrument",
        on_delete=models.CASCADE,
        related_name="preferred_classification_portfolio",
        limit_choices_to=(models.Q(instrument_type__is_classifiable=True) & models.Q(level=0)),
    )
    classification = models.ForeignKey(
        "wbfdm.Classification",
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        related_name="preferred_classification_throughs",
    )
    classification_group = models.ForeignKey(
        "wbfdm.ClassificationGroup",
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        related_name="preferred_classification_group_throughs",
    )

    def __str__(self) -> str:
        return f"{self.portfolio} - {self.instrument}: ({self.classification})"

    def save(self, *args, **kwargs) -> None:
        if not self.classification_group and self.classification:
            self.classification_group = self.classification.group
        return super().save(*args, **kwargs)

    class Meta:
        verbose_name = "Portfolio Prefered Classification Per Instrument"
        verbose_name_plural = "Portfolio Prefered Classification Per Instruments"
        constraints = [
            models.UniqueConstraint(
                fields=["portfolio", "instrument", "classification_group"],
                name="unique_prefered_classification_relationship",
            ),
        ]


class PortfolioTradableExchangeThroughModelManager(models.Manager):
    def get_for_adjusment_factors(self, portfolio: "Portfolio", exchange: "Exchange") -> tuple[float, float]:
        try:
            rel = PortfolioTradableExchangeThroughModel.objects.get(portfolio=portfolio, exchange=exchange)
            return rel.rebalancing_adjustment_factor, rel.exposure_adjustment_factor
        except PortfolioTradableExchangeThroughModel.DoesNotExist:
            return 0.0, 0.0


class PortfolioTradableExchangeThroughModel(models.Model):
    portfolio = models.ForeignKey(
        "wbportfolio.Portfolio", on_delete=models.CASCADE, related_name="tradable_exchange_relationships"
    )
    exchange = models.ForeignKey(
        "wbfdm.Exchange",
        on_delete=models.CASCADE,
        related_name="tradable_portfolio_relationships",
    )
    rebalancing_adjustment_factor = models.FloatField(default=0)
    exposure_adjustment_factor = models.FloatField(default=0)

    objects = PortfolioTradableExchangeThroughModelManager()

    def __str__(self):
        return f"Adjustment Factor for {self.portfolio} and {self.exchange}"

    class Meta:
        verbose_name = "Portfolio Tradeable Exchange"
        verbose_name_plural = "Portfolio Tradeable Exchanges"
        constraints = [
            models.UniqueConstraint(
                fields=["portfolio", "exchange"],
                name="unique_tradable_exchange_per_portfolio",
            ),
        ]


@receiver(pre_merge, sender="wbfdm.Instrument")
def pre_merge_instrument(sender: models.Model, merged_object: "Instrument", main_object: "Instrument", **kwargs):
    """
    Reassign all merged instrument preferred classification relationship to the main instrument
    """
    for through in PortfolioInstrumentPreferredClassificationThroughModel.objects.filter(instrument=merged_object):
        if classification := through.classification:
            PortfolioInstrumentPreferredClassificationThroughModel.objects.get_or_create(
                portfolio=through.portfolio,
                instrument=main_object,
                classification_group=(
                    through.classification_group if through.classification_group else classification.group
                ),
                defaults={"classification": classification},
            )
        through.delete()



from collections import defaultdict
from typing import Optional

from wbcore.enums import Unit
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class ESGMetricAggregationPortfolioPandasDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = [
            dp.Field(
                key=None,
                label="Portfolio Data",
                children=[
                    dp.Field(
                        key="underlying_instrument_repr",
                        label="Name",
                        width=Unit.PIXEL(300),
                    ),
                    dp.Field(key="weighting", label="Weight", width=Unit.PIXEL(100)),
                    dp.Field(key="total_value_fx_usd", label="Total Asset (USD)", width=Unit.PIXEL(150)),
                ],
            )
        ]
        esg_data_fields = [
            dp.Field(
                key="esg_data",
                label=self.view.esg_aggregation.get_esg_code().name,
                width=Unit.PIXEL(150),
            )
        ]
        for extra in self.view.dataloader.extra_esg_data_logs:
            esg_data_fields.append(dp.Field(key=extra.series.name, label=extra.label, width=Unit.PIXEL(150)))
        fields.append(
            dp.Field(key=None, label="ESG Data", children=esg_data_fields),
        )
        if self.view.dataloader.intermediary_logs:
            intermediary_fields = []
            intermediary_groups = defaultdict(list)

            for intermediary in self.view.dataloader.intermediary_logs:
                field = dp.Field(
                    key=intermediary.series.name,
                    label=intermediary.label,
                    width=Unit.PIXEL(100),
                )
                if intermediary.group:
                    intermediary_groups[intermediary.group].append(field)

                else:
                    intermediary_fields.append(
                        dp.Field(
                            key=intermediary.series.name,
                            label=intermediary.label,
                            width=Unit.PIXEL(150),
                        )
                    )
            for group_name, intermediary_groups_fields in intermediary_groups.items():
                intermediary_fields.append(dp.Field(key=None, label=group_name, children=intermediary_groups_fields))
            fields.append(
                dp.Field(key=None, label="Intermediary", children=intermediary_fields),
            )
        fields.append(
            dp.Field(
                key=None,
                label="Metric",
                children=[
                    dp.Field(key="metric", label=self.view.esg_aggregation.value, width=Unit.PIXEL(150)),
                    dp.Field(
                        key="weights_in_coverage",
                        label="Weight in Coverage",
                        width=Unit.PIXEL(100),
                    ),
                ],
            )
        )
        return dp.ListDisplay(fields=fields)



# Generated by Django 5.0.9 on 2024-11-11 11:59

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0070_remove_assetposition_unique_asset_position_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="trade",
            options={"ordering": ("order",), "verbose_name": "Trade", "verbose_name_plural": "Trades"},
        ),
        migrations.AlterField(
            model_name="trade",
            name="order",
            field=models.PositiveIntegerField(db_index=True, default=0, editable=False, verbose_name="order"),
            preserve_default=False,
        ),
    ]



from django.contrib import admin

from wbportfolio.models import DividendTransaction


@admin.register(DividendTransaction)
class DividendAdmin(admin.ModelAdmin):
    search_fields = ["portfolio__name", "underlying_instrument__computed_str"]
    list_filter = ("distribution_method", "portfolio")
    list_display = (
        "distribution_method",
        "portfolio",
        "underlying_instrument",
        "value_date",
        "currency",
        "currency_fx_rate",
        "total_value",
        "total_value_fx_portfolio",
    )
    fieldsets = (
        (
            "Dividend Information",
            {
                "fields": (
                    (
                        "portfolio",
                        "underlying_instrument",
                        "import_source",
                    ),
                    ("ex_date", "record_date", "value_date"),
                    ("currency", "currency_fx_rate"),
                    ("price", "shares", "retrocession"),
                    ("total_value", "total_value_gross"),
                    ("total_value_fx_portfolio", "total_value_gross_fx_portfolio"),
                    ("created", "updated"),
                    ("comment",),
                )
            },
        ),
    )
    readonly_fields = [
        "total_value",
        "total_value_gross",
        "total_value_fx_portfolio",
        "total_value_gross_fx_portfolio",
        "created",
        "updated",
    ]
    autocomplete_fields = ["portfolio", "underlying_instrument", "currency"]
    ordering = ("-value_date",)
    raw_id_fields = ["import_source"]



from decimal import Decimal
from unittest.mock import patch

import pytest
from pandas._libs.tslibs.offsets import BDay

from wbportfolio.models import AssetPosition, Order, OrderProposal, Portfolio
from wbportfolio.models.adjuster import adjust_quote
from wbportfolio.models.asset import adjust_assets
from wbportfolio.models.orders.orders import adjust_orders
from wbportfolio.models.utils import get_adjusted_shares


def test_get_adjusted_shares():
    assert get_adjusted_shares(Decimal("150"), Decimal("100"), Decimal("200")) == Decimal("75")


@pytest.mark.django_db
@patch.object(Order, "_get_price")
def test_adjust_orders(mock_get_price, weekday, order_factory, instrument_factory):
    mock_get_price.return_value = (Decimal("100"), Decimal("0"))
    old_quote = instrument_factory.create()
    new_quote = instrument_factory.create()
    o1 = order_factory.create(
        underlying_instrument=old_quote, order_proposal__trade_date=weekday, price=Decimal("100"), shares=Decimal("10")
    )
    o2 = order_factory.create(
        underlying_instrument=old_quote, order_proposal__trade_date=weekday, price=Decimal("100"), shares=Decimal("20")
    )
    adjust_orders(Order.objects.filter(underlying_instrument=old_quote), new_quote)

    o1.refresh_from_db()
    o2.refresh_from_db()

    assert o1.underlying_instrument == new_quote
    assert o2.underlying_instrument == new_quote
    assert o1.price == Decimal("100")
    assert o2.price == Decimal("100")
    assert o1.shares == Decimal("10")
    assert o2.shares == Decimal("20")

    mock_get_price.return_value = (Decimal("200"), Decimal("0"))
    adjust_orders(Order.objects.filter(underlying_instrument=new_quote), old_quote)

    o1.refresh_from_db()
    o2.refresh_from_db()

    assert o1.underlying_instrument == old_quote
    assert o2.underlying_instrument == old_quote
    assert o1.price == Decimal("200")
    assert o2.price == Decimal("200")
    assert o1.shares == Decimal("5")
    assert o2.shares == Decimal("10")


@pytest.mark.django_db
def test_adjust_assets(weekday, asset_position_factory, instrument_factory, instrument_price_factory):
    old_quote = instrument_factory.create()
    new_quote = instrument_factory.create()
    old_quote_price = instrument_price_factory.create(
        instrument=old_quote, net_value=Decimal("100"), date=weekday, calculated=False
    )
    new_quote_price = instrument_price_factory.create(
        instrument=new_quote, net_value=Decimal("100"), date=weekday, calculated=False
    )

    a1 = asset_position_factory.create(
        underlying_quote=old_quote, date=weekday, initial_price=Decimal("100"), initial_shares=Decimal("10")
    )
    a2 = asset_position_factory.create(
        underlying_quote=old_quote, date=weekday, initial_price=Decimal("100"), initial_shares=Decimal("20")
    )
    adjust_assets(AssetPosition.objects.filter(underlying_quote=old_quote), new_quote)

    a1.refresh_from_db()
    a2.refresh_from_db()

    assert a1.underlying_quote == new_quote
    assert a2.underlying_quote == new_quote
    assert a1.underlying_quote_price == new_quote_price
    assert a2.underlying_quote_price == new_quote_price
    assert a1.initial_price == Decimal("100")
    assert a2.initial_price == Decimal("100")
    assert a1.initial_shares == Decimal("10")
    assert a2.initial_shares == Decimal("20")

    old_quote_price.net_value = new_quote_price.net_value = Decimal("200")
    old_quote_price.save()
    new_quote_price.save()

    adjust_assets(AssetPosition.objects.filter(underlying_quote=new_quote), old_quote)

    a1.refresh_from_db()
    a2.refresh_from_db()

    assert a1.underlying_quote == old_quote
    assert a2.underlying_quote == old_quote
    assert a1.underlying_quote_price == old_quote_price
    assert a2.underlying_quote_price == old_quote_price
    assert a1.initial_price == Decimal("200")
    assert a2.initial_price == Decimal("200")
    assert a1.initial_shares == Decimal("5")
    assert a2.initial_shares == Decimal("10")


@pytest.mark.django_db
@patch("wbportfolio.models.adjuster.adjust_assets")
@patch("wbportfolio.models.adjuster.adjust_orders")
@patch.object(OrderProposal, "replay")
def test_adjust_quote(mock_replay, mock_adjust_orders, mock_adjust_assets, weekday, order_factory, instrument_factory):
    old_quote = instrument_factory.create()
    new_quote = instrument_factory.create()
    o1 = order_factory.create(  # noqa: F841
        underlying_instrument=old_quote, order_proposal__trade_date=weekday, price=Decimal("100"), shares=Decimal("10")
    )
    o2 = order_factory.create(  # noqa: F841
        underlying_instrument=old_quote,
        order_proposal__trade_date=(weekday + BDay(1)),
        price=Decimal("100"),
        shares=Decimal("10"),
    )
    o3 = order_factory.create(
        underlying_instrument=old_quote,
        order_proposal__trade_date=(weekday + BDay(1)),
        price=Decimal("100"),
        shares=Decimal("10"),
    )
    adjust_quote(
        old_quote,
        new_quote,
        adjust_after=weekday,
        only_portfolios=Portfolio.objects.filter(id=o3.order_proposal.portfolio.id),
    )

    mock_adjust_assets.asser_called_once()
    assert set(mock_adjust_assets.call_args[0][0]) == set(AssetPosition.objects.none())
    assert mock_adjust_assets.call_args[0][1] == new_quote

    mock_adjust_orders.assert_called_once()
    assert set(mock_adjust_orders.call_args[0][0]) == set(Order.objects.filter(id=o3.id))
    assert mock_adjust_orders.call_args[0][1] == new_quote

    mock_replay.assert_called_once_with(reapply_order_proposal=True)



from wbcore.menus import ItemPermission, MenuItem

VALUATIONDATAAUM_MENUITEM = MenuItem(
    label="AuM Chart",
    endpoint="wbportfolio:aumchart-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal, permissions=["wbportfolio.view_portfolio"]
    ),
)



# Generated by Django 5.2.9 on 2026-02-04 08:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0096_portfoliotradableexchangethroughmodel'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='product',
            options={'permissions': [('administrate_product', 'Can administrate Product')], 'verbose_name': 'Product', 'verbose_name_plural': 'Products'},
        ),
        migrations.AddField(
            model_name='product',
            name='next_termination_reminder_date',
            field=models.DateField(blank=True, null=True),
        ),
    ]



import datetime
import re
from io import BytesIO

import numpy as np
import pandas as pd

from wbportfolio.import_export.utils import get_file_extension


def file_name_parse(file_name):
    isin = re.findall("([A-Z]{2}[A-Z0-9]{9}[0-9]{1})", file_name)

    if len(isin) != 1:
        raise ValueError("Not exactly 1 isin found in the filename")

    return {
        "isin": isin[0],
    }


def parse(import_source):
    if get_file_extension(import_source.file) == ".xlsx":
        df = pd.read_excel(BytesIO(import_source.file.read()), engine="openpyxl", sheet_name="Summary").dropna(
            how="all"
        )
    elif get_file_extension(import_source.file) == ".xls":
        df = pd.read_excel(BytesIO(import_source.file.read()), sheet_name="Summary").dropna(how="all")
    else:
        raise Exception("File not supported")

    parts = file_name_parse(import_source.file.name)
    isin = parts["isin"]

    xx, yy = np.where(df == "Instrument")
    df = df.iloc[: xx[0], :].transpose().dropna(how="all", axis=0).reset_index(drop=True)

    df = df.rename(columns=df.iloc[0]).drop(df.index[0])

    date = df.loc[:, df.columns.str.contains("Statement as at Close of Business")].dropna(how="any").iloc[0, 0]
    date = datetime.datetime.strptime(date, "%Y-%m-%d").date()
    value = df.loc[:, df.columns.str.contains("RPL")].dropna(how="any").iloc[0, 0]

    data = [
        {
            "instrument": {"instrument_type": "product", "isin": isin},
            "date": date.strftime("%Y-%m-%d"),
            "net_value": value,
            "calculated": False,
        }
    ]

    return {"data": data}



from datetime import date
from decimal import Decimal

import pandas as pd
import pytest
from django.core.exceptions import ValidationError
from pandas._libs.tslibs.offsets import BDay

from wbportfolio.models import AssetPosition, OrderProposal


@pytest.mark.django_db
class TestRebalancingModel:
    def test_get_target_portfolio(
        self, rebalancing_model, portfolio, weekday, asset_position_factory, instrument_price_factory
    ):
        trade_date = (weekday + BDay(1)).date()
        with pytest.raises(ValidationError):  # trigger value error because rebalancing not valid (no position yet)
            rebalancing_model.get_target_portfolio(portfolio, trade_date, weekday)
        a1 = asset_position_factory(weighting=0.7, portfolio=portfolio, date=weekday)
        a2 = asset_position_factory(weighting=0.3, portfolio=portfolio, date=weekday)
        instrument_price_factory.create(instrument=a1.underlying_quote, date=trade_date)
        target_portfolio = rebalancing_model.get_target_portfolio(portfolio, (weekday + BDay(1)).date(), weekday)
        target_positions = target_portfolio.positions_map
        assert target_positions[a1.underlying_instrument.id].weighting == 0.5
        assert target_positions[a2.underlying_instrument.id].weighting == 0.5


@pytest.mark.django_db
class TestRebalancer:
    def test_activation_date(self, rebalancer_factory, asset_position):
        r1 = rebalancer_factory.create(portfolio=asset_position.portfolio)
        assert r1.activation_date == asset_position.date

        r2 = rebalancer_factory.create()
        r2.activation_date = date.today()

    def test_is_valid(self, weekday, rebalancer_factory):
        # default frequency is monthly

        first_of_the_month = weekday.replace(day=1)
        rebalancer = rebalancer_factory.create(frequency="RRULE:FREQ=MONTHLY;", activation_date=first_of_the_month)
        assert rebalancer.is_valid(first_of_the_month)
        if weekday.day != 1:
            assert not rebalancer.is_valid(weekday)

    def test_evaluate_rebalancing(
        self, weekday, rebalancer_factory, asset_position_factory, instrument_factory, instrument_price_factory
    ):
        rebalancer = rebalancer_factory.create(apply_order_proposal_automatically=True)
        trade_date = (weekday + BDay(1)).date()

        i1 = instrument_factory.create()
        i2 = instrument_factory.create()
        instrument_price_factory.create(instrument=i1, date=weekday)
        instrument_price_factory.create(instrument=i1, date=trade_date)
        instrument_price_factory.create(instrument=i2, date=weekday)
        instrument_price_factory.create(instrument=i2, date=trade_date)

        a1 = asset_position_factory.create(
            portfolio=rebalancer.portfolio, date=weekday, weighting=0.7, underlying_instrument=i1
        )
        a2 = asset_position_factory.create(
            portfolio=rebalancer.portfolio, date=weekday, weighting=0.3, underlying_instrument=i2
        )
        order_proposal = rebalancer.evaluate_rebalancing(trade_date)
        assert order_proposal.orders.exclude(underlying_instrument__is_cash=True).count() == 2
        assert order_proposal.status == OrderProposal.Status.CONFIRMED
        assert AssetPosition.objects.get(
            portfolio=rebalancer.portfolio, date=trade_date, underlying_quote=a1.underlying_instrument
        ).weighting == Decimal(0.5)
        assert AssetPosition.objects.get(
            portfolio=rebalancer.portfolio, date=trade_date, underlying_quote=a2.underlying_instrument
        ).weighting == Decimal(0.5)

    def test_get_rrule(self, weekday, rebalancer_factory):
        # default frequency is monthly

        first_of_the_month = weekday.replace(day=1)
        until = first_of_the_month.replace(year=first_of_the_month.year + 1)
        rebalancer = rebalancer_factory.create(frequency="RRULE:FREQ=MONTHLY;", activation_date=first_of_the_month)

        valid_dates = set(rebalancer.get_rrule(until))
        assert valid_dates
        assert set(valid_dates) == set(pd.date_range(start=first_of_the_month, end=until, freq="MS"))



import datetime
from datetime import timedelta
from decimal import Decimal
from unittest.mock import call, patch

import pandas as pd
import pytest
from django.contrib.auth.models import Permission
from dynamic_preferences.registries import global_preferences_registry
from faker import Faker
from wbfdm.models import Instrument

from wbportfolio.models import Product, Trade

from ...models.utils import get_termination_message
from .utils import PortfolioTestMixin

fake = Faker()


@pytest.mark.django_db
class TestProductModel(PortfolioTestMixin):
    def test_init(self, product):
        assert product.id is not None

    def test_white_label(self, white_label_product):
        assert white_label_product.white_label_product

    def test_total_value(self, product_factory, trade_factory, instrument_price_factory):
        product = product_factory.create()
        t1 = trade_factory.create(
            shares=100,
            underlying_instrument=product,
            transaction_date=datetime.date(2010, 1, 1) - datetime.timedelta(days=2),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        t2 = trade_factory.create(
            shares=50,
            underlying_instrument=product,
            transaction_date=datetime.date(2010, 1, 1) - datetime.timedelta(days=1),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        t3 = trade_factory.create(  # noise
            shares=10,
            underlying_instrument=product,
            transaction_date=datetime.date(2010, 1, 1),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        p1 = instrument_price_factory.create(
            instrument=Instrument.objects.get(id=product.id), date=t1.transaction_date
        )
        p2 = instrument_price_factory.create(
            instrument=Instrument.objects.get(id=product.id), date=t2.transaction_date
        )
        p3 = instrument_price_factory.create(
            instrument=Instrument.objects.get(id=product.id), date=t3.transaction_date
        )
        assert product.get_total_aum_usd(p1.date) == pytest.approx(Decimal(0), rel=Decimal("1e-4"))
        assert product.get_total_aum_usd(p2.date) == pytest.approx(
            Decimal(p2.net_value * (t1.shares)), rel=Decimal("1e-4")
        )
        assert product.get_total_aum_usd(p3.date) == pytest.approx(
            Decimal(p3.net_value * (t1.shares + t2.shares)), rel=Decimal("1e-4")
        )

    def test_total_shares_with_date(self, product_factory, trade_factory):
        date1 = datetime.date(2010, 1, 1)
        date2 = datetime.date(2010, 1, 2)
        product = product_factory.create()
        sub1 = trade_factory.create(
            underlying_instrument=product,
            transaction_date=date1 - datetime.timedelta(days=1),
            value_date=date1,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        sub2 = trade_factory.create(
            underlying_instrument=product,
            transaction_date=date2 - datetime.timedelta(days=1),
            value_date=date2,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        red1 = trade_factory.create(
            underlying_instrument=product,
            transaction_date=date1 - datetime.timedelta(days=1),
            value_date=date1,
            transaction_subtype=Trade.Type.REDEMPTION,
        )
        red2 = trade_factory.create(
            underlying_instrument=product,
            transaction_date=date2 - datetime.timedelta(days=1),
            value_date=date2,
            transaction_subtype=Trade.Type.REDEMPTION,
        )
        assert product.total_shares(date1 - datetime.timedelta(days=1)) == 0
        assert product.total_shares(date1) == red1.shares + sub1.shares
        assert product.total_shares(date2) == red1.shares + red2.shares + sub1.shares + sub2.shares

    def test_nominal_value_with_date(self, product_factory, trade_factory):
        product = product_factory.create()
        trade_factory.create_batch(
            5,
            shares=100,
            underlying_instrument=product,
            transaction_date=datetime.date(2010, 1, 1) - datetime.timedelta(days=1),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        trade_factory.create_batch(
            5,
            shares=100,
            underlying_instrument=product,
            transaction_date=datetime.date(2010, 1, 2) - datetime.timedelta(days=1),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        assert product.nominal_value(datetime.date(2010, 1, 1)) == (500 * product.share_price)
        assert product.nominal_value(datetime.date(2010, 1, 2)) == (1000 * product.share_price)

    def test_get_title(self, product, user):
        assert product.get_title()

    def test_compute_str(self, product, user):
        assert product.compute_str()

        # test if we save the parent instrument, the computed str is the one from the product class
        instrument = product.instrument_ptr
        instrument.save()
        product.refresh_from_db()
        assert product.computed_str == product.compute_str()

    def test_urlify_title(self, product):
        assert product.name.lower() in product.urlify_title

    def test_get_products_internal_user(self, product, white_label_product, internal_user_factory):
        assert set(Product.get_products(internal_user_factory.create().profile)) == {product, white_label_product}

    def test_get_products_normal_user(self, product_factory, user, company):
        public_product = product_factory.create()
        assert set(Product.get_products(user.profile)) == {public_product}

        white_label_product_profile = product_factory.create()
        white_label_product_profile.white_label_customers.add(user.profile)

        white_label_product_employer = product_factory.create()
        user.profile.employers.add(company)
        white_label_product_employer.white_label_customers.add(company)

        assert set(Product.get_products(user.profile)) == {
            white_label_product_employer,
            white_label_product_profile,
            public_product,
        }

    def test_subquery_is_white_label_product(self, product, white_label_product):
        product_queryset = Product.objects.all().annotate(is_white_label=Product.subquery_is_white_label_product())
        assert product_queryset.filter(is_white_label=True).count() == 1

    def test_annotate_aum(
        self, portfolio, product_factory, instrument_price_factory, trade_factory, currency_fx_rates_factory
    ):
        p1 = product_factory.create()
        d = datetime.date(2021, 1, 1)
        c1 = trade_factory.create(
            portfolio=p1.portfolio,
            underlying_instrument=p1,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
            value_date=d,
            transaction_date=d,
        )
        v1 = instrument_price_factory.create(instrument=p1, date=d, outstanding_shares=c1.shares)
        fx_rate = p1.currency.fx_rates.get(date=d)
        qs = Product.annotate_last_aum(Product.objects.all(), val_date=d)
        p1.update_last_valuation_date()
        assert float(qs.first().assets_under_management_usd) == pytest.approx(
            float(v1.net_value) * float(c1.shares) / float(fx_rate.value)
        )

    def test_account_merging(self, account_factory, product):
        base_account = account_factory.create()
        merged_account = account_factory.create()
        product.default_account = merged_account
        product.save()
        base_account.merge(merged_account)
        product.refresh_from_db()
        assert product.default_account == base_account

    def test_high_water_mark(self, product, instrument_price_factory):
        initial_price = instrument_price_factory.create(instrument=product, net_value=100)
        p1 = instrument_price_factory.create(instrument=product, net_value=200)
        p2 = instrument_price_factory.create(instrument=product, net_value=100)
        p3 = instrument_price_factory.create(instrument=product, net_value=300)
        p4 = instrument_price_factory.create(instrument=product, net_value=100)
        assert product.get_high_water_mark(p1.date) == initial_price.net_value
        assert product.get_high_water_mark(p2.date) == p1.net_value
        assert product.get_high_water_mark(p3.date) == p1.net_value
        assert product.get_high_water_mark(p4.date) == p3.net_value
        product.initial_high_water_mark = 400
        product.save()
        assert product.get_high_water_mark(p4.date) == 400

    @patch("wbportfolio.models.products.send_notification")
    def test_check_and_notify_product_termination_on_date(self, mock_send_notification, product, internal_user):
        global_preferences = global_preferences_registry.manager()
        global_preferences["wbportfolio__product_termination_notice_period"] = 60
        delisted_date = datetime.date(2026, 2, 4)
        default_reminder_date = datetime.date(2025, 12, 6)

        product.delisted_date = delisted_date
        product.save()

        assert product.next_termination_reminder_date == default_reminder_date

        for _d in pd.date_range(default_reminder_date, delisted_date + timedelta(days=5)):
            product.check_and_notify_product_termination_on_date(_d.date())

        mock_send_notification.assert_not_called()  # nobody has the administator permission yet

        # reset reminder date
        product.next_termination_reminder_date = None
        product.save()
        internal_user.user_permissions.add(Permission.objects.get(codename="administrate_product"))

        for _d in pd.date_range(default_reminder_date, delisted_date + timedelta(days=5)):
            product.check_and_notify_product_termination_on_date(_d.date())

        expected_calls = [
            call(
                code="wbportfolio.product.termination_notice",
                title="Product Termination Notice",
                body=get_termination_message(product, datetime.date(2025, 12, 6)),  # 60 days in advance
                user=internal_user,
            ),
            call(
                code="wbportfolio.product.termination_notice",
                title="Product Termination Notice",
                body=get_termination_message(product, datetime.date(2026, 1, 5)),  # 60/2=30 days in advance
                user=internal_user,
            ),
            call(
                code="wbportfolio.product.termination_notice",
                title="Product Termination Notice",
                body=get_termination_message(product, datetime.date(2026, 1, 20)),  # int(30/2) = 15
                user=internal_user,
            ),
            call(
                code="wbportfolio.product.termination_notice",
                title="Product Termination Notice",
                body=get_termination_message(product, datetime.date(2026, 1, 28)),  # int(15/2) = 7  days in advance
                user=internal_user,
            ),
            call(
                code="wbportfolio.product.termination_notice",
                title="Product Termination Notice",
                body=get_termination_message(
                    product, datetime.date(2026, 2, 2)
                ),  # int(7/2) = 3 but 2  days in advance because it falls within a weekend day
                user=internal_user,
            ),
            call(
                code="wbportfolio.product.termination_notice",
                title="Product Termination Notice",
                body=get_termination_message(product, datetime.date(2026, 2, 3)),  # int(2/2) = 1 day in advance
                user=internal_user,
            ),
        ]
        mock_send_notification.assert_has_calls(expected_calls)

    def test_delist_product_disable_report(self, product, report_factory):
        report = report_factory.create(content_object=product, is_active=True)
        assert product.delisted_date is None
        assert report.is_active

        product.delisted_date = datetime.date.today()
        product.save()

        report.refresh_from_db()
        assert report.is_active is False



from django.db.models import Exists, OuterRef
from wbcore import filters as wb_filters

from wbportfolio.models import InstrumentPortfolioThroughModel, OrderProposal


class OrderProposalFilterSet(wb_filters.FilterSet):
    has_custodian_adapter = wb_filters.BooleanFilter(
        method="filter_has_custodian_adapter", label="Has Custodian Adapter"
    )
    waiting_for_input = wb_filters.BooleanFilter(method="filter_waiting_for_input", label="Waiting for Input")
    is_automatic_rebalancing = wb_filters.BooleanFilter(
        method="filter_is_automatic_rebalancing", label="Automatic Rebalancing"
    )

    def filter_has_custodian_adapter(self, queryset, name, value):
        queryset = queryset.annotate(
            has_custodian_adapter=Exists(
                InstrumentPortfolioThroughModel.objects.filter(
                    portfolio=OuterRef("portfolio"), instrument__net_asset_value_computation_method_path__isnull=False
                )
            )
        )
        if value is True:
            queryset = queryset.filter(has_custodian_adapter=True)
        elif value is False:
            queryset = queryset.filter(has_custodian_adapter=False)
        return queryset

    def filter_waiting_for_input(self, queryset, name, value):
        input_status = [OrderProposal.Status.PENDING, OrderProposal.Status.DRAFT, OrderProposal.Status.APPROVED]
        if value is True:
            queryset = queryset.filter(status__in=input_status)
        elif value is False:
            queryset = queryset.exclude(status__in=input_status)
        return queryset

    def filter_is_automatic_rebalancing(self, queryset, name, value):
        if value is True:
            queryset = queryset.filter(rebalancing_model__isnull=False)
        elif value is False:
            queryset = queryset.filter(rebalancing_model__isnull=True)
        return queryset

    class Meta:
        model = OrderProposal
        fields = {
            "trade_date": ["exact"],
            "status": ["exact"],
            "rebalancing_model": ["exact"],
            "portfolio": ["exact"],
            "creator": ["exact"],
            "approver": ["exact"],
            "execution_status": ["exact"],
        }



import csv
from io import StringIO

import numpy as np
import pandas as pd
from xlrd import xldate_as_datetime

from wbportfolio.models import Product, Trade

from .sylk import SYLK
from .utils import get_portfolio_id

MAP = {
    "register_id": "register__register_reference",
    "register_name": "bank",
    "isin": "underlying_instrument",
    "trans_date": "transaction_date",
    "settl_date": "value_date",
    "trans_ref": "external_id",
    "trans_price": "price",
    "quantity": "shares",
    "note": "comment",
}


def parse(import_source):
    data = list()
    sylk_handler = SYLK()
    merger_isin_map = import_source.source.import_parameters.get("custom_merged_isin_map", {"XX": "LU"})
    merger_export_parts_trans_desc = import_source.source.import_parameters.get(
        "merger_export_parts_trans_desc", "Export Parts"
    )

    for line in [_line.decode("cp1252") for _line in import_source.file.open("rb").readlines()]:
        sylk_handler.parseline(line)

    buffer = StringIO()
    csvwriter = csv.writer(buffer, quotechar="'", delimiter=";", lineterminator="\n", quoting=csv.QUOTE_ALL)
    for line in sylk_handler.stream_rows():
        csvwriter.writerow(line)

    buffer.seek(0)
    content = buffer.read().replace('""', "")
    df = pd.read_csv(StringIO(content), sep=";", quotechar="'")
    if not df.empty:
        # Filter out all non transaction rows and remove record_desc col
        df = df[df["record_desc"].str.strip() == "TRANSACTION"]

        # Convert timestamps to json conform date strings
        df["trans_date"] = df["trans_date"].apply(lambda x: xldate_as_datetime(x, datemode=0).date())
        df["trans_date"] = df["trans_date"].apply(lambda x: x.strftime("%Y-%m-%d"))
        df["settl_date"] = df["settl_date"].apply(lambda x: xldate_as_datetime(x, datemode=0).date())
        df["settl_date"] = df["settl_date"].apply(lambda x: x.strftime("%Y-%m-%d"))

        # Replace all nan values with empty str
        df[["register_firstname", "cust_ref"]] = df[["register_firstname", "cust_ref"]].replace(np.nan, "", regex=True)

        # Merge register_firstname and cust_ref
        df["note"] = df["register_firstname"] + df["cust_ref"]
        # the trade marked with "Export Parts" and where isin is marked for merge need to be filter out as they are introduced by the merger
        df = df[
            ~(
                (df["trans_desc"] == merger_export_parts_trans_desc)
                & (df["isin"].str.contains("|".join(merger_isin_map.keys())))
            )
        ]
        for pat, repl in merger_isin_map.items():
            df["isin"] = df["isin"].str.replace(pat, repl)

        # Create Product Mapping and apply to df
        product_mapping = {
            product["isin"]: product["id"]
            for product in Product.objects.filter(isin__in=df["isin"].unique()).values("id", "isin")
        }
        df["isin"] = df["isin"].apply(lambda x: product_mapping[x])

        # Rename Columns
        df = df.rename(columns=MAP)
        df = df[df.columns.intersection(MAP.values())]

        df["transaction_subtype"] = df.shares.apply(
            lambda x: Trade.Type.REDEMPTION.value if x < 0 else Trade.Type.SUBSCRIPTION.value
        )
        df["portfolio"] = df["underlying_instrument"].apply(lambda x: get_portfolio_id(x))
        df["underlying_instrument"] = df["underlying_instrument"].apply(
            lambda x: {"id": x, "instrument_type": "product"}
        )
        df["pending"] = False
        # Convert df to list of dicts

        data = df.to_dict("records")

    return {
        "data": data,
    }



from django.contrib import admin

from wbportfolio.models.orders import Order


class OrderTabularInline(admin.TabularInline):
    model = Order
    fk_name = "order_proposal"

    readonly_fields = [
        "_effective_weight",
        "_target_weight",
        "_effective_shares",
        "_target_shares",
        "total_value",
        "total_value_gross",
        "total_value_fx_portfolio",
        "total_value_gross_fx_portfolio",
        "created",
        "updated",
    ]

    fields = [
        "underlying_instrument",
        "_effective_weight",
        "_target_weight",
        "weighting",
        "shares",
        "daily_return",
    ]

    raw_id_fields = ["underlying_instrument"]



import random

import pytest
from faker import Faker
from psycopg.types.range import NumericRange
from wbcompliance.factories.risk_management import RuleThresholdFactory

from wbportfolio.risk_management.backends.exposure_portfolio import (
    RuleBackend as ExposurePortfolioRuleBackend,
)
from wbportfolio.tests.models.utils import PortfolioTestMixin

fake = Faker()


@pytest.mark.django_db
class TestExposurePortfolioRuleModel(PortfolioTestMixin):
    @pytest.mark.parametrize(
        "group_by",
        [groupby for groupby in ExposurePortfolioRuleBackend.GroupbyChoices.values],
    )
    def test_check_rule_groupby_weighting(
        self,
        weekday,
        group_by,
        instrument_factory,
        cash_factory,
        asset_position_factory,
        instrument_price_factory,
        country_factory,
        currency_factory,
        instrument_type_factory,
        portfolio,
        classification,
    ):
        only_type = instrument_type_factory.create()
        only_country = country_factory.create()
        only_currency = currency_factory.create()
        parameters = {
            "group_by": group_by,
            "field": ExposurePortfolioRuleBackend.Field.WEIGHTING.value,
            "asset_classes": [only_type.id],
            "countries": [only_country.id],
            "currencies": [only_currency.id],
            "classifications": [classification.id],
        }

        lower = random.random()
        upper = random.uniform(lower, 1)
        exposure_portfolio_backend = ExposurePortfolioRuleBackend(
            weekday,
            portfolio,
            parameters,
            [RuleThresholdFactory.create(range=NumericRange(lower=lower, upper=upper))],  # type: ignore
        )

        i1 = instrument_factory.create(
            country=only_country, currency=only_currency, instrument_type=only_type, classifications=[classification]
        )

        threshold = exposure_portfolio_backend.thresholds[0]
        instrument_price_factory.create(instrument=i1, date=weekday)
        asset_position_factory.create(
            date=weekday,
            weighting=random.uniform(threshold.range.lower, threshold.range.upper),
            underlying_instrument=i1,
            portfolio=portfolio,
        )  # Breached position

        asset_position_factory.create(
            date=weekday,
            weighting=random.uniform(threshold.range.lower, threshold.range.upper),
            portfolio=portfolio,
        )  # Breached position but filtered out

        i2 = cash_factory.create()
        instrument_price_factory.create(instrument=i2, date=weekday)
        asset_position_factory.create(
            date=weekday,
            weighting=random.uniform(0, threshold.range.lower),
            underlying_instrument=i2,
            portfolio=portfolio,
        )  # None breached position

        incidents = list(exposure_portfolio_backend.check_rule())
        assert len(incidents) == 1
        if group_by == ExposurePortfolioRuleBackend.GroupbyChoices.CURRENCY:
            assert incidents[0].breached_object == only_currency
        elif group_by == ExposurePortfolioRuleBackend.GroupbyChoices.COUNTRY:
            assert incidents[0].breached_object == only_country
        elif group_by in [
            ExposurePortfolioRuleBackend.GroupbyChoices.PRIMARY_CLASSIFICATION,
            ExposurePortfolioRuleBackend.GroupbyChoices.FAVORITE_CLASSIFICATION,
        ]:
            assert incidents[0].breached_object == classification

    def test_exposure_rule_with_extra_filter_field(
        self, weekday, portfolio, asset_position_factory, instrument_price_factory, instrument_factory
    ):
        parameters = {
            "group_by": "instrument_type",
            "field": ExposurePortfolioRuleBackend.Field.WEIGHTING.value,
            "extra_filter_field": "market_capitalization_usd",
            "extra_filter_field_lower_bound": 1000.0,
            "extra_filter_field_upper_bound": 10000.0,
        }
        exposure_portfolio_backend = ExposurePortfolioRuleBackend(
            weekday,
            portfolio,
            parameters,
            [RuleThresholdFactory.create(range=NumericRange(lower=0.05, upper=None))],  # type: ignore
        )
        i1 = instrument_factory.create()
        i2 = instrument_factory.create(instrument_type=i1.instrument_type)
        i3 = instrument_factory.create(instrument_type=i1.instrument_type)
        instrument_price_factory.create(instrument=i1, date=weekday, market_capitalization=1000)
        instrument_price_factory.create(instrument=i2, date=weekday, market_capitalization=9999)
        instrument_price_factory.create(instrument=i3, date=weekday, market_capitalization=10001)
        asset_position_factory.create(
            date=weekday,
            weighting=0.025,
            underlying_instrument=i1,
            portfolio=portfolio,
        )
        asset_position_factory.create(
            date=weekday,
            weighting=0.025,
            underlying_instrument=i2,
            portfolio=portfolio,
        )

        asset_position_factory.create(
            date=weekday,
            weighting=0.95,
            underlying_instrument=i3,
            portfolio=portfolio,
        )
        incidents = list(exposure_portfolio_backend.check_rule())
        assert len(incidents) == 1
        incident = incidents[0]
        assert incident.breached_object_repr == i1.instrument_type.name
        assert incident.breached_value == '<span style="color:green">+5.00%</span>'



from typing import Generator

from wbcompliance.models.risk_management import backend
from wbcompliance.models.risk_management.dispatch import register

from wbportfolio.pms.typing import Valuation as ValuationDTO

from .mixins import ActiveProductRelationshipMixin, StopLossMixin


@register("Stop Loss Instrument Rule Backend", rule_group_key="portfolio")
class RuleBackend(ActiveProductRelationshipMixin, StopLossMixin, backend.AbstractRuleBackend):
    def is_passive_evaluation_valid(self) -> bool:
        return (
            super().is_passive_evaluation_valid()
            and self.product
            and self.product.valuations.filter(date=self.evaluation_date).exists()
        )

    def _process_dto(
        self, instrument_valuation_dto: ValuationDTO, benchmark_valuation_dto: ValuationDTO = None, *args, **kwargs
    ) -> Generator[backend.IncidentResult, None, None]:
        perf_instrument = self._get_performance(instrument_valuation_dto)
        perf_benchmark = self._get_performance(benchmark_valuation_dto)
        yield from self._generate_incidents(self.product.id, perf_instrument, perf_benchmark)



# Generated by Django 5.0.12 on 2025-11-28 13:11

from django.db import migrations, models

def handle_data(apps, schema_editor):
    PortfolioPortfolioThroughModel = apps.get_model('wbportfolio', 'PortfolioPortfolioThroughModel')
    PortfolioPortfolioThroughModel.objects.filter(type="PRIMARY").update(type="LOOK_THROUGH")
    from wbportfolio.models.portfolio import Portfolio
    for portfolio in Portfolio.objects.all():
        if portfolio.assets.exists():
            val_date = portfolio.assets.latest("date").date
            for parent_ptf, _ in portfolio.get_parent_portfolios(val_date):
                PortfolioPortfolioThroughModel.objects.get_or_create(portfolio_id=portfolio.id, dependency_portfolio_id=parent_ptf.id, defaults={"type": "HIERARCHICAL"})
class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0092_order_quantization_error_alter_orderproposal_status'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='portfolioportfoliothroughmodel',
            name='unique_primary',
        ),
        migrations.AlterField(
            model_name='portfolioportfoliothroughmodel',
            name='type',
            field=models.CharField(choices=[('LOOK_THROUGH', 'Look-through'), ('MODEL', 'Model'), ('CUSTODIAN', 'Custodian'), ('HIERARCHICAL', 'Hierarchical')], default='LOOK_THROUGH', verbose_name='Type'),
        ),
        migrations.AddConstraint(
            model_name='portfolioportfoliothroughmodel',
            constraint=models.UniqueConstraint(condition=models.Q(('type', 'LOOK_THROUGH')), fields=('portfolio', 'type'), name='unique_lookthrough'),
        ),
        migrations.RunPython(handle_data)
    ]



import random
from datetime import timedelta
from decimal import Decimal

import factory
from faker import Faker

from wbportfolio.models import Trade

fake = Faker()


class TradeFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Trade

    currency_fx_rate = Decimal(1.0)
    fees = Decimal(0.0)
    portfolio = factory.SubFactory("wbportfolio.factories.PortfolioFactory")
    underlying_instrument = factory.SubFactory("wbfdm.factories.InstrumentFactory")
    currency = factory.SubFactory("wbcore.contrib.currency.factories.CurrencyFactory")
    transaction_date = factory.Faker("date_object")
    value_date = factory.LazyAttribute(lambda o: o.transaction_date + timedelta(days=1))
    bank = factory.Faker("company")
    marked_for_deletion = False
    shares = factory.Faker("pydecimal", min_value=10, max_value=1000, right_digits=4)
    price = factory.LazyAttribute(lambda o: random.randint(10, 10000))
    # order_proposal = factory.SubFactory("wbportfolio.factories.OrderProposalFactory")


class CustomerTradeFactory(TradeFactory):
    transaction_subtype = factory.LazyAttribute(
        lambda o: Trade.Type.REDEMPTION if o.shares < 0 else Trade.Type.SUBSCRIPTION
    )
    underlying_instrument = factory.SubFactory("wbportfolio.factories.ProductFactory")
    portfolio = factory.LazyAttribute(lambda x: x.underlying_instrument.primary_portfolio)



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class AssetPositionEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None


class AssetPositionPortfolioEndpointConfig(AssetPositionEndpointConfig):
    pass


class AssetPositionEquityEndpointConfig(AssetPositionEndpointConfig):
    pass


class AssetPositionInstrumentEndpointConfig(AssetPositionEndpointConfig):
    pass


class AssetPositionIndexEndpointConfig(AssetPositionEndpointConfig):
    pass


class AssetPositionProductGroupEndpointConfig(AssetPositionEndpointConfig):
    pass


class CashPositionPortfolioEndpointConfig(AssetPositionEndpointConfig):
    PK_FIELD = "portfolio"

    def get_instance_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:portfolio-list",
            args=[],
            request=self.request,
        )

    def get_update_endpoint(self, **kwargs):
        return None


class ContributorPortfolioChartEndpointConfig(AssetPositionEndpointConfig):
    pass


class AssetPositionUnderlyingInstrumentChartEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:instrument-assetpositionchart-list",
            args=[self.view.kwargs["instrument_id"]],
            request=self.request,
        )


class DistributionChartEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:portfolio-distributionchart-list",
            args=[self.view.kwargs["portfolio_id"]],
            request=self.request,
        )


class DistributionTableEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None


class CompositionModelPortfolioPandasEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None



from decimal import Decimal

from wbportfolio.pms.typing import Portfolio, Position
from wbportfolio.rebalancing.base import AbstractRebalancingModel
from wbportfolio.rebalancing.decorators import register
from wbportfolio.rebalancing.models.mixins import InstrumentMixin


@register("Equally Weighted Rebalancing")
class EquallyWeightedRebalancing(InstrumentMixin, AbstractRebalancingModel):
    def __init__(self, *args, **kwargs):
        super(EquallyWeightedRebalancing, self).__init__(*args, **kwargs)
        if not self.effective_portfolio:
            self.effective_portfolio = self.portfolio._build_dto(self.trade_date)
        self.instruments = self.get_instruments(**kwargs)
        self.prices, self.returns_df = self.instruments.get_returns_df(
            from_date=self.last_effective_date, to_date=self.trade_date, use_dl=True
        )
        try:
            self.prices = self.prices[self.trade_date]
        except KeyError:
            self.prices = {}
        try:
            self.returns_df = self.returns_df.loc[self.trade_date, :].to_dict()
        except KeyError:
            self.returns_df = {}

    def is_valid(self) -> bool:
        return self.instruments.exists() and self.prices

    def get_target_portfolio(self) -> Portfolio:
        positions = []
        instrument_count = len(self.instruments)
        for underlying_instrument in self.instruments:
            positions.append(
                Position(
                    date=self.trade_date,
                    asset_valuation_date=self.trade_date,
                    underlying_instrument=underlying_instrument.id,
                    weighting=Decimal(1 / instrument_count),
                    price=self.prices.get(underlying_instrument.id, None),
                    daily_return=self.returns_df.get(underlying_instrument.id, Decimal("0")),
                )
            )
        return Portfolio(positions)



import math
from datetime import date, timedelta
from decimal import Decimal

from pandas._libs.tslibs.offsets import BDay

from wbportfolio.preferences import get_product_termination_notice_period


def get_next_termination_reminder_businessdate(delisted_date: date, today: date) -> date | None:
    """Computes next business day reminder using geometric interval from delisted_date.

    Args:
        delisted_date: Product termination date.
        today: Current date.

    Returns:
        Next reminder business date or delisted_date if imminent.
    """
    remaining_days = (delisted_date - today).days
    initial_period_days = get_product_termination_notice_period()
    if remaining_days <= 1:
        return delisted_date + timedelta(days=1)

    # Geometric step: 2^(log2(ratio)-1)
    interval = pow(2, int(math.log2(remaining_days / initial_period_days)) - 1)
    next_termination_reminder_date = delisted_date - timedelta(days=int(initial_period_days * interval))

    # Weekend → next business day
    return (next_termination_reminder_date - timedelta(days=1) + BDay(1)).date()


def get_termination_message(product, today: date) -> str:
    day_until_termination = (product.delisted_date - today).days
    humanized_days = "tomorrow" if day_until_termination == 1 else f"{day_until_termination} days"
    return f"The product {product} will be terminated in {humanized_days} ({product.delisted_date:%Y-%m-%d}). Please act accordingly."


def get_adjusted_shares(old_shares: Decimal, old_price: Decimal, new_price: Decimal) -> Decimal:
    return old_shares * (old_price / new_price)



from datetime import date

from django.db import models

MUTUAL_FUND_TIMEDELTA_DAY_SHIFT = 7


class DataBackendMixin:
    DATE_LABEL = "valuation_date"

    def is_object_valid(self, obj: models.Model) -> bool:
        return super().is_object_valid(obj) and obj.assets.exists() and obj.is_active_at_date(date.today())

    def get_provider_id(self, obj: models.Model) -> str:
        return obj.id



# Generated by Django 5.2.9 on 2025-12-09 08:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbfdm', '0035_alter_instrument_related_instruments_and_more'),
        ('wbportfolio', '0093_remove_portfolioportfoliothroughmodel_unique_primary_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='portfolio',
            name='depends_on',
            field=models.ManyToManyField(blank=True, related_name='dependent_portfolios', through='wbportfolio.PortfolioPortfolioThroughModel', through_fields=('portfolio', 'dependency_portfolio'), to='wbportfolio.portfolio', verbose_name='The portfolios this portfolio depends on'),
        ),
        migrations.AlterField(
            model_name='portfolio',
            name='preferred_instrument_classifications',
            field=models.ManyToManyField(blank=True, limit_choices_to=models.Q(('instrument_type__is_classifiable', True), ('level', 0)), related_name='preferred_portfolio_classifications', through='wbportfolio.PortfolioInstrumentPreferredClassificationThroughModel', through_fields=('portfolio', 'instrument'), to='wbfdm.instrument', verbose_name='The Preferred classification per instrument'),
        ),
    ]
